@@ -1,5 +1,7 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+
+from collections import defaultdict
 from decimal import Decimal
 from itertools import groupby
 
@@ -8,15 +10,16 @@ from sql.aggregate import Sum
 from sql.conditionals import Case, Coalesce
 from sql.functions import Abs
 
-from trytond import backend
 from trytond.i18n import gettext
 from trytond.model import ModelSQL, ModelView, fields
+from trytond.modules.account.exceptions import (
+    CancelWarning, DelegateLineWarning, GroupLineWarning,
+    RescheduleLineWarning)
 from trytond.modules.company.model import CompanyValueMixin
 from trytond.modules.currency.fields import Monetary
 from trytond.pool import Pool, PoolMeta
 from trytond.pyson import Bool, Eval, Id, If
-from trytond.tools.multivalue import migrate_property
-from trytond.transaction import Transaction
+from trytond.transaction import Transaction, check_access
 from trytond.wizard import (
     Button, StateAction, StateTransition, StateView, Wizard)
 
@@ -27,7 +30,7 @@ from .payment import KINDS
 class MoveLine(metaclass=PoolMeta):
     __name__ = 'account.move.line'
     payment_amount = fields.Function(Monetary(
-            "Payment Amount",
+            "Amount to Pay",
             currency='payment_currency', digits='payment_currency',
             states={
                 'invisible': ~Eval('payment_kind'),
@@ -58,16 +61,23 @@ class MoveLine(metaclass=PoolMeta):
         super(MoveLine, cls).__setup__()
         cls._buttons.update({
                 'pay': {
-                    'invisible': ~Eval('payment_kind').in_(
-                        list(dict(KINDS).keys())),
+                    'invisible': (
+                        ~Eval('payment_kind').in_(list(dict(KINDS).keys()))
+                        | Eval('reconciliation')),
                     'depends': ['payment_kind'],
                     },
                 'payment_block': {
-                    'invisible': Eval('payment_blocked', False),
+                    'invisible': (
+                        ~Eval('payment_kind').in_(list(dict(KINDS).keys()))
+                        | Eval('reconciliation')
+                        | Eval('payment_blocked', False)),
                     'depends': ['payment_blocked'],
                     },
                 'payment_unblock': {
-                    'invisible': ~Eval('payment_blocked', False),
+                    'invisible': (
+                        ~Eval('payment_kind').in_(list(dict(KINDS).keys()))
+                        | Eval('reconciliation')
+                        | ~Eval('payment_blocked', False)),
                     'depends': ['payment_blocked'],
                     },
                 })
@@ -200,6 +210,7 @@ class MoveLine(metaclass=PoolMeta):
                 ('credit', '<', 0),
                 ],
             ('maturity_date', '<=', date),
+            ('payment_blocked', '!=', True),
             ]
 
     @classmethod
@@ -210,7 +221,7 @@ class MoveLine(metaclass=PoolMeta):
         Reception = pool.get('party.party.reception_direct_debit')
         if date is None:
             date = Date.today()
-        with Transaction().set_context(_check_access=True):
+        with check_access():
             lines = cls.search(cls._pay_direct_debit_domain(date))
 
         payments = []
@@ -475,24 +486,6 @@ class ConfigurationPaymentGroupSequence(ModelSQL, CompanyValueMixin):
                 Id('account_payment', 'sequence_type_account_payment_group')),
             ])
 
-    @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-
-        super(ConfigurationPaymentGroupSequence, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('payment_group_sequence')
-        value_names.append('payment_group_sequence')
-        fields.append('company')
-        migrate_property(
-            'account.configuration', field_names, cls, value_names,
-            fields=fields)
-
     @classmethod
     def default_payment_group_sequence(cls):
         pool = Pool()
@@ -504,6 +497,104 @@ class ConfigurationPaymentGroupSequence(ModelSQL, CompanyValueMixin):
             return None
 
 
+class MoveCancel(metaclass=PoolMeta):
+    __name__ = 'account.move.cancel'
+
+    def transition_cancel(self):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        moves_w_payments = []
+        for move in self.records:
+            for line in move.lines:
+                if any(p.state != 'failed' for p in line.payments):
+                    moves_w_payments.append(move)
+                    break
+        if moves_w_payments:
+            names = ', '.join(
+                m.rec_name for m in moves_w_payments[:5])
+            if len(moves_w_payments) > 5:
+                names += '...'
+            key = Warning.format('cancel_payments', moves_w_payments)
+            if Warning.check(key):
+                raise CancelWarning(
+                    key, gettext(
+                        'account_payment.msg_move_cancel_payments',
+                        moves=names))
+        return super().transition_cancel()
+
+
+class MoveLineGroup(metaclass=PoolMeta):
+    __name__ = 'account.move.line.group'
+
+    def do_group(self, action):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        lines_w_payments = []
+        for line in self.records:
+            if any(p.state != 'failed' for p in line.payments):
+                lines_w_payments.append(line)
+        if lines_w_payments:
+            names = ', '.join(
+                m.rec_name for m in lines_w_payments[:5])
+            if len(lines_w_payments) > 5:
+                names += '...'
+            key = Warning.format('group_payments', lines_w_payments)
+            if Warning.check(key):
+                raise GroupLineWarning(
+                    key, gettext(
+                        'account_payment.msg_move_line_group_payments',
+                        lines=names))
+        return super().do_group(action)
+
+
+class MoveLineReschedule(metaclass=PoolMeta):
+    __name__ = 'account.move.line.reschedule'
+
+    def do_reschedule(self, action):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        lines_w_payments = []
+        for line in self.records:
+            if any(p.state != 'failed' for p in line.payments):
+                lines_w_payments.append(line)
+        if lines_w_payments:
+            names = ', '.join(
+                m.rec_name for m in lines_w_payments[:5])
+            if len(lines_w_payments) > 5:
+                names += '...'
+            key = Warning.format('reschedule_payments', lines_w_payments)
+            if Warning.check(key):
+                raise RescheduleLineWarning(
+                    key, gettext(
+                        'account_payment.msg_move_line_reschedule_payments',
+                        lines=names))
+        return super().do_reschedule(action)
+
+
+class MoveLineDelegate(metaclass=PoolMeta):
+    __name__ = 'account.move.line.delegate'
+
+    def do_delegate(self, action):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        lines_w_payments = []
+        for line in self.records:
+            if any(p.state != 'failed' for p in line.payments):
+                lines_w_payments.append(line)
+        if lines_w_payments:
+            names = ', '.join(
+                m.rec_name for m in lines_w_payments[:5])
+            if len(lines_w_payments) > 5:
+                names += '...'
+            key = Warning.format('delegate_payments', lines_w_payments)
+            if Warning.check(key):
+                raise DelegateLineWarning(
+                    key, gettext(
+                        'account_payment.msg_move_line_delegate_payments',
+                        lines=names))
+        return super().do_delegate(action)
+
+
 class Invoice(metaclass=PoolMeta):
     __name__ = 'account.invoice'
 
@@ -561,6 +652,43 @@ class Invoice(metaclass=PoolMeta):
         return amounts
 
 
+class Statement(metaclass=PoolMeta):
+    __name__ = 'account.statement'
+
+    @classmethod
+    def create_move(cls, statements):
+        moves = super().create_move(statements)
+        cls._process_payments(moves)
+        return moves
+
+    @classmethod
+    def _process_payments(cls, moves):
+        pool = Pool()
+        Payment = pool.get('account.payment')
+
+        to_success = defaultdict(set)
+        to_fail = defaultdict(set)
+        for move, statement, lines in moves:
+            for line in lines:
+                for kind, payments in line.payments():
+                    if (kind == 'receivable') == (line.amount >= 0):
+                        to_success[line.date].update(payments)
+                    else:
+                        to_fail[line.date].update(payments)
+        # The failing should be done last because success is usually not a
+        # definitive state
+        if to_success:
+            for date, payments in to_success.items():
+                with Transaction().set_context(clearing_date=date):
+                    Payment.succeed(Payment.browse(payments))
+        if to_fail:
+            for date, payments in to_fail.items():
+                with Transaction().set_context(clearing_date=date):
+                    Payment.fail(Payment.browse(payments))
+        payments = set.union(*to_success.values(), *to_fail.values())
+        return list(payments)
+
+
 class StatementLine(metaclass=PoolMeta):
     __name__ = 'account.statement.line'
 
@@ -614,6 +742,11 @@ class StatementLine(metaclass=PoolMeta):
             if self.payment.line:
                 self.account = self.payment.line.account
 
+    def payments(self):
+        "Yield payments per kind"
+        if self.payment:
+            yield self.payment.kind, [self.payment]
+
 
 class Dunning(metaclass=PoolMeta):
     __name__ = 'account.dunning'
@@ -623,9 +756,9 @@ class Dunning(metaclass=PoolMeta):
 
     @classmethod
     def search_active(cls, name, clause):
-        if tuple(clause[1:]) in {('=', True), ('!=', False)}:
+        if tuple(clause[1:]) in [('=', True), ('!=', False)]:
             domain = ('line.payment_amount', '>', 0)
-        elif tuple(clause[1:]) in {('=', False), ('!=', True)}:
+        elif tuple(clause[1:]) in [('=', False), ('!=', True)]:
             domain = ('line.payment_amount', '<=', 0)
         else:
             domain = []
